# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

import os
import zipfile
import hashlib
import json
import requests
import shutil
from appdirs import *
from sapphire.common import util

PUBLISHED_AT_FILENAME = 'published_at.txt'

class FirmwarePackageDirNotFound(Exception):
    pass

def data_dir():
    return user_data_dir("Chromatron", "SapphireOpenSystems")

def firmware_package_dir(release=None):
    package_dir = os.path.join(data_dir(), 'firmware_packages')

    if release:
        return os.path.join(package_dir, release)

    else:
        return package_dir

def open_manifest(dir):
    with open('manifest.txt', 'r') as f:
        return json.loads(f.read())

def get_firmware(release=None, include_firmware_image=False, sort_fwid=False):
    cwd = os.getcwd()

    try:
        os.chdir(firmware_package_dir(release=release))

    except OSError:
        raise FirmwarePackageDirNotFound(firmware_package_dir())

    firmwares = {}

    # scan firmware zips
    for filename in os.listdir(os.getcwd()):
        filepath = os.path.join(os.getcwd(), filename)
        
        try:

            filedir, ext = os.path.splitext(filename)
                
            if ext != '.zip':
                continue

            try:
                os.mkdir(filedir)

            except OSError:
                pass

            zf = zipfile.ZipFile(filename)
            zf.extractall(filedir)
            zf.close()

            os.chdir(filedir)

            # open the manifest
            manifest = open_manifest(os.getcwd())

            key = filedir
            if sort_fwid:
                key = str(manifest['fwid']).replace('-', '')

            firmwares[key] = {'manifest': manifest,
                              'short_name': filedir}

            if include_firmware_image:
                # load binary
                try:
                    with open('firmware.bin', 'rb') as f:
                        image = f.read()

                except IOError:
                    with open('wifi_firmware.bin', 'rb') as f:
                        image = f.read()

                firmwares[key]['image'] = {'binary': image}

                # verify firmware
                fw_sha256 = hashlib.sha256(image).hexdigest()
                valid = (fw_sha256 == firmwares[key]['manifest']['sha256'])
                firmwares[key]['image']['valid'] = valid

                if not valid:
                    # if not valid, delete the binary image, so
                    # we can't accidentally load it
                    del firmwares[key]['image']['binary']


            os.chdir('..')

            # clean up directory
            shutil.rmtree(filedir, True)


        except IOError:
            pass

        except zipfile.BadZipfile:
            pass


    # change back
    os.chdir(cwd)

    return firmwares

def get_releases(use_date_for_key=False):
    cwd = os.getcwd()
    releases = {}

    try:
        for filename in os.listdir(firmware_package_dir()):    
            os.chdir(firmware_package_dir())

            if not os.path.isdir(filename):
                continue

            os.chdir(filename)
            with open(PUBLISHED_AT_FILENAME, 'r') as f:
                # published_at = util.iso_to_datetime(f.read())
                published_at = f.read()

            if use_date_for_key:
                releases[published_at] = filename

            else:
                releases[filename] = published_at

    except OSError:
        # firmware dir does not exist
        return {}

    # change back
    os.chdir(cwd)

    return releases


def get_most_recent_release():
    releases = get_releases(use_date_for_key=True)
    
    most_recent = sorted(releases.keys())[-1]

    return releases[most_recent]

def update_releases():
    # make sure the package dir exists
    try:
        os.makedirs(firmware_package_dir())

    except OSError:
        pass
    
    # get local releases
    local_releases = get_releases()

    # get releases
    r = requests.get('https://api.github.com/repos/sapphireos/chromatron/releases')

    if r.status_code != 200:
        return []

    # sort releases
    releases = [a for a in reversed(sorted(r.json(), key=lambda a: a['published_at']))]

    new_releases = []

    for release in releases:
        # check if we already have this release
        if release['tag_name'] in local_releases:
            continue
        
        # we don't have it, let's download it
        for asset in [a for a in release['assets'] if a['content_type'] == 'application/zip']:
            # print asset['name'], asset['url'], asset['size']

            # get zip file
            headers = {'Accept': 'application/octet-stream'}
            r = requests.get(asset['url'], headers=headers)

            filename = os.path.join(firmware_package_dir(), asset['name'])
            with open(filename, 'wb') as f: # must be in binary mode or Windows will break
                f.write(r.content)

            # unzip it
            filedir = os.path.join(firmware_package_dir(), release['tag_name'])

            zf = zipfile.ZipFile(filename)
            zf.extractall(filedir)
            # for z in zf.infolist():
            #     print z.filename
            zf.close()

            # I *really* hate Mac sometimes.

            macosx_dir = os.path.join(filedir, "__MACOSX")
            try:
                shutil.rmtree(macosx_dir)
            except OSError:
                pass

        new_releases.append(release['tag_name'])

    return new_releases

if __name__ == "__main__":
    from pprint import pprint

    # pprint(get_firmware())
    pprint(get_releases(use_date_for_key=True))
    # pprint(get_most_recent_release())
    # update_releases()







#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

from elysianfields import *
from sapphire.fields.protocol import *
from sapphiredata import *

CMD2_APP_CMD_BASE = 32768


class DeviceCommandProtocol(Protocol):

    PORT = 16385

    class Echo(Payload):
        msg_type = 1
        fields = [StringField(_length=64, _name="echo_data")]

    class Reboot(Payload):
        msg_type = 2
        fields = []

    class SafeMode(Payload):
        msg_type = 3
        fields = []

    class LoadFirmware(Payload):
        msg_type = 4
        fields = []

    class FormatFS(Payload):
        msg_type = 10
        fields = []

    class GetFileID(Payload):
        msg_type = 20
        fields = [StringField(_length=64, _name="name")]

    class CreateFile(Payload):
        msg_type = 21
        fields = [StringField(_length=64, _name="name")]

    class ReadFileData(Payload):
        msg_type = 22
        fields = [Uint8Field(_name="file_id"),
                  Uint32Field(_name="position"),
                  Uint32Field(_name="length")]

    class WriteFileData(Payload):
        msg_type = 23
        fields = [Uint8Field(_name="file_id"),
                  Uint32Field(_name="position"),
                  Uint32Field(_name="length"),
                  RawBinField(_name="data")]

    class RemoveFile(Payload):
        msg_type = 24
        fields = [Uint8Field(_name="file_id")]

    class ResetCfg(Payload):
        msg_type = 32
        fields = []

    class SetRegbits(Payload):
        msg_type = 40
        fields = [Uint16Field(_name="reg"),
                  Uint8Field(_name="mask")]

    class ClrRegbits(Payload):
        msg_type = 41
        fields = [Uint16Field(_name="reg"),
                  Uint8Field(_name="mask")]

    class GetRegbits(Payload):
        msg_type = 42
        fields = [Uint16Field(_name="reg")]

    class SetKV(Payload):
        msg_type = 80
        fields = [RawBinField(_name="data")]

    class GetKV(Payload):
        msg_type = 81
        fields = [RawBinField(_name="data")]

    class SetKVServer(Payload):
        msg_type = 85
        fields = [Ipv4Field(_name="ip"),
                  Uint16Field(_name="port")]

    class SetSecurityKey(Payload):
        msg_type = 90
        fields = [Uint8Field(_name="key_id"),
                  Key128Field(_name="key")]

    msg_type_format = Uint16Field()


class DeviceCommandResponseProtocol(Protocol):

    class Echo(Payload):
        msg_type = 1
        fields = [StringField(_length=64, _name="echo_data")]

    class Reboot(Payload):
        msg_type = 2
        fields = []

    class SafeMode(Payload):
        msg_type = 3
        fields = []

    class LoadFirmware(Payload):
        msg_type = 4
        fields = []

    class FormatFS(Payload):
        msg_type = 10
        fields = []

    class GetFileID(Payload):
        msg_type = 20
        fields = [Int8Field(_name="file_id")]

    class CreateFile(Payload):
        msg_type = 21
        fields = [Int8Field(_name="file_id")]

    class ReadFileData(Payload):
        msg_type = 22
        fields = [RawBinField(_name="data")]

    class WriteFileData(Payload):
        msg_type = 23
        fields = [Uint16Field(_name="write_length")]

    class RemoveFile(Payload):
        msg_type = 24
        fields = [Uint8Field(_name="status")]

    class ResetCfg(Payload):
        msg_type = 32
        fields = []

    class SetRegbits(Payload):
        msg_type = 40
        fields = []

    class ClrRegbits(Payload):
        msg_type = 41
        fields = []

    class GetRegbits(Payload):
        msg_type = 42
        fields = [Uint8Field(_name="data")]

    class SetKV(Payload):
        msg_type = 80
        fields = [RawBinField(_name="data")]

    class GetKV(Payload):
        msg_type = 81
        fields = [RawBinField(_name="data")]

    class SetKVServer(Payload):
        msg_type = 85
        fields = []

    class SetSecurityKey(Payload):
        msg_type = 90
        fields = []

    msg_type_format = Uint16Field()


if __name__ == '__main__':
    pass



#! python
#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

import threading

from sapphire.devices.device import Device, DeviceUnreachableException

import sys
import traceback # ignore unused warning!

import cmd2 as cmd

from pyparsing import *


# set up query parse grammar
query_item   = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\()*+,-./:;<>?@[\\]^_`{|}~'
equals       = Suppress('=')
query_param  = Word(query_item).setResultsName('param')
query_value  = Word(query_item).setResultsName('value')
query_expr   = Group(query_param + equals + query_value)
query_list   = ZeroOrMore(query_expr)

def make_query_dict(line):
    q = query_list.parseString(line)

    d = dict()

    for i in q:
        d[i.param] = i.value

    return d

class SapphireConsole(cmd.Cmd):

    prompt = '(Nothing): '

    ruler = '-'

    def __init__(self, targets=[], devices=[]):
        self.targets = targets
        self.devices = devices

        cmd.Cmd.__init__(self)

    def cmdloop(self):
        SapphireConsole.next_cmd = None

        cmd.Cmd.cmdloop(self)

        return SapphireConsole.next_cmd

    def perror(self, errmsg, statement=None):
        traceback.print_exc()
        print (str(errmsg))

    def do_scan(self, line):
        pass

    def query(self, line):
        if line == 'all':
            devices = self.devices

        else:
            qdict = make_query_dict(line)
            devices = [o for o in self.devices if o.query(**qdict)]

        return devices

    def init_shell(self, query):
        # set up next command shell
        if len(self.targets) > 0:
            target_type = type(self.targets[0])

            # check if all targets are the same type
            for target in self.targets:
                if type(target) != target_type:
                    target_type = Device
                    break

            SapphireConsole.next_cmd = makeConsole(targets=self.targets,
                                                   devices=self.devices,
                                                   device=target_type())

            #SapphireConsole.next_cmd.prompt = "(%s): " % (query)
            SapphireConsole.next_cmd.prompt = "(%3d devices): " % (len(self.targets))

        else:
            # no targets, set up empty shell
            SapphireConsole.next_cmd = SapphireConsole()

            SapphireConsole.next_cmd.prompt = "(Nothing): "

    def do_query(self, line):
        devices = self.query(line)

        if devices:
            print "Found %d devices" % (len(devices))

            for device in devices:
                print device.who()

    def do_select(self, line):
        # query for targets
        devices = self.query(line)

        self.targets = devices

        self.init_shell(line)

        print "Selected %d devices" % (len(self.targets))

        return True

    def do_add(self, line):
        # query for targets
        devices = self.query(line)

        for device in devices:
            self.targets.append(device)

        self.init_shell(line)

        print "Selected %d devices" % (len(self.targets))

        return True

    def do_remove(self, line):
        # query for targets
        devices = self.query(line)

        for device in devices:
            self.targets.remove(device)

        self.init_shell(line)

        print "Selected %d devices" % (len(self.targets))

        return True

    def do_who(self, line):
        for target in self.targets:
            print target.who()


cli_template = """
    def do_$fname(self, line):
        for target in self.targets:
            sys.stdout.write('%s@%s: ' % (target.name.ljust(16), target.host))

            try:
                print target.cli_$fname(line)

            except DeviceUnreachableException as e:
                print 'Error:%s from %s' % (e, target.host)

            except Exception as e:
                print 'Error: %s' % (e)
                #traceback.print_exc()

"""

def makeConsole(targets=[], devices=[], device=None):
    if not device:
        try:
            device = targets[0]
        except:
            pass

    # get command line function listing from device
    cli_funcs = device.get_cli()

    s = "class aConsole(SapphireConsole):\n"

    for fname in cli_funcs:
        s += cli_template.replace('$fname', fname)

    exec(s)

    return aConsole(targets=targets, devices=devices)

def main():
    try:
        host = sys.argv[1]
        sys.argv[1] = '' # prevents an unknown syntax error in the command loop

        try:
            print "Connecting to %s" % host

            d = Device(host=host)
            d.scan()

            c = makeConsole(targets=[d])
            c.cmdloop()

        except:
            print "*** Unable to connect to %s" % host
            raise

    except IndexError:
        print "No device specified"


if __name__ == '__main__':
    main()


# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

from setuptools import setup

setup(
    name='sapphire',

    version='1.0.2',

    packages=['sapphire',
              'sapphire.fields',
              'sapphire.common',
              'sapphire.query',
              'sapphire.udpx',
              'sapphire.devices',
              'sapphire.buildtools',
              ],

    package_data={'sapphire.buildtools': ['settings.json', 'linker.x', 'project_template/*']},

    license='GNU General Public License v3',

    description='Sapphire Tools',

    long_description=open('README.txt').read(),

    install_requires=[
        "catbus",
        "appdirs==1.4.3",
        "requests==2.18.4",
        "setuptools == 34.2.0",
        "pyserial == 3.2.1",
        "bitstring == 3.1.5",
        "cmd2 == 0.6.9",
        "click >= 6.6",
        "crcmod == 1.7",
        "pyparsing == 2.2.0",
        "intelhex == 2.1",
        "pydispatcher == 2.0.5",
        "elysianfields >= 1.0",
        "fnvhash",
    ],

    entry_points='''
        [console_scripts]
        sapphiremake=sapphire.buildtools.core:main
        sapphireconsole=sapphire.devices.sapphireconsole:main
    ''',
)



#
# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>
#

"""UDPX socket module.

UDPX is a thin protocol on top of UDP that provides acknowledged datagram
delivery with an automatic repeat request mechanism in the client.


"""


import socket
import random
import time

import bitstring



class Packet(object):

    VERSION = 0
    HEADER_FORMAT = 'uint:2, uint:1, uint:1, uint:1, uint:3, uint:8'

    def __init__(self,
                 server=False,
                 ack_request=True,
                 ack=False,
                 id=None,
                 data=''):

        self.__version  = Packet.VERSION

        self.__server = 0
        self.__ack_request = 0
        self.__ack = 0

        if server:
            self.__server = 1

        if ack_request:
            self.__ack_request = 1

        if ack:
            self.__ack = 1

        if id is None:
            id = random.randint(0,255)

        self.__id = id

        self.__payload = data

        self.time = None

    def __str__(self):

        s = "Ver:%d | Svr:%d | Arq:%d | Ack:%d | ID:%3d | PayloadLength:%3d" % \
            (self.__version,
             self.__server,
             self.__ack_request,
             self.__ack,
             self.__id,
             len(self.__payload))

        if self.time:
            s += " | Time(ms):%5d" % (self.time * 1000)

        return s

    def pack(self):
        header = bitstring.pack(Packet.HEADER_FORMAT,
                                self.__version,
                                self.__server,
                                self.__ack_request,
                                self.__ack,
                                0,
                                self.__id)

        return header.bytes + self.__payload

    def unpack(self, data):
        s = bitstring.BitStream(bytes=data[0:2])
        header = s.unpack(Packet.HEADER_FORMAT)

        self.__version      = header[0]
        self.__server       = header[1]
        self.__ack_request  = header[2]
        self.__ack          = header[3]
        # header[4] is reserved bits
        self.__id           = header[5]

        self.__payload = data[2:]

        return self

    def get_version(self):
        return self.__version

    def get_server(self):
        return self.__server != 0

    def get_ack_request(self):
        return self.__ack_request != 0

    def get_ack(self):
        return self.__ack != 0

    def get_id(self):
        return self.__id

    def get_payload(self):
        return self.__payload

    def set_payload(self, value):
        self.__payload = value

    version = property(get_version)
    server = property(get_server)
    ack_request = property(get_ack_request)
    ack = property(get_ack)
    id = property(get_id)
    payload = property(get_payload, set_payload)


class ClientSocket(object):

    DEFAULT_TRIES = 10
    INITIAL_TIMEOUT = 1.0
    MIN_TIMEOUT = 0.25

    def __init__(self, addr_family=None, sock_type=None):
        self._sock = socket.socket(socket.AF_INET,
                                    socket.SOCK_DGRAM)

        self._min_timeout = ClientSocket.MIN_TIMEOUT
        self._tries = ClientSocket.DEFAULT_TRIES
        self._initial_timeout = ClientSocket.INITIAL_TIMEOUT

        self._received_data = None
        self._received_addr = None

        self._packet_id = random.randint(0,255)

        self.rtt = -1
        self.filter = 0.1
        self.timeout_multiplier = 3.0

    def flush(self):
        self._received_data = None
        self._received_addr = None

        # set sock to non-blocking
        self._sock.setblocking(0)
        try:
            while True:
                ack_data, host = self._sock.recvfrom(4096)

        except socket.error:
            pass

        # set back to blocking
        self._sock.setblocking(1)

    def bind(self, address):
        self._sock.bind(address)

    def settimeout(self, seconds):
        self._initial_timeout = seconds

    def setsockopt(self, level, optname, value):
        self._sock.setsockopt(level, optname, value)

    def sendto(self, data, address):

        # Flush socket.
        # This is really important.  If we bail out from sendto() due to some error condition
        # and the socket then receives a packet, the next time we call sendto() we will receive
        # the previous buffered packet (from the cancelled transaction).  This will often cause
        # a retry because the ack packet will have the wrong ID from the packet we just sent.
        # However, you can't rely on that, because the ID space is only 8 bits, so an ack packet
        # with a matching ID but incorrect data may still get through.
        #
        self.flush()

        # cannot broadcast on a client socket
        assert address[0] != '<broadcast>'

        self._packet_id += 1
        self._packet_id %= 256
        
        self._sock.connect(address)

        # build data packet
        packet = Packet(data=data, id=self._packet_id)

        start = time.time()

        # print '',
        # retry loop
        for i in xrange(self._tries):

            # set initial timeout
            timeout = self._initial_timeout
            # timeout = 2.0

            # set timeout
            # print i, timeout
            self._sock.settimeout(timeout)

            rtt_start = time.time()

            # send packet
            try:
                self._sock.send(packet.pack())

            except socket.error:
                # we'll get this if the host is unreachable,
                # or if some other error occurred
                raise


            # wait for timeout or received data
            try:
                ack_data, host = self._sock.recvfrom(4096)

                # parse ack
                ack_msg = Packet().unpack(ack_data)

                # print packet
                # print ack_msg, host
                # print hex(ord(ack_data[0])), hex(ord(ack_data[1]))

                # check packet for errors
                if ack_msg.version != ack_msg.VERSION:
                    raise InvalidPacketException(packet)

                elif not ack_msg.server:
                    raise InvalidPacketException(packet)

                elif ack_msg.ack_request:
                    raise InvalidPacketException(packet)

                elif not ack_msg.ack:
                    raise InvalidPacketException(packet)

                ack_msg.time = time.time() - start
                rtt = time.time() - rtt_start

                # initialize RTT
                if self.rtt < 0:
                    self.rtt = rtt

                else:
                    # update filtered rtt
                    self.rtt = (self.filter * rtt) + ((1.0 - self.filter) * self.rtt)

                    # update timeout
                    self._initial_timeout = self.timeout_multiplier * self.rtt

                if self._initial_timeout < self._min_timeout:
                    self._initial_timeout = self._min_timeout

                # print rtt, self.rtt, self._initial_timeout
                # print host, ack_msg

                if ack_msg.id != packet.id:
                    # print "Incorrect msg ID %x != %x" % (ack_msg.id, packet.id)
                    # raise InvalidPacketException(packet)   

                    self._initial_timeout += 1.0

                    continue

                # set data and server host buffers
                self._received_data = ack_msg.payload
                self._received_addr = host

                # print "message completed on try: %d" % (i)

                return

            except socket.timeout:

                # print "timeout"

                # increase timeout, exponential backoff
                timeout = timeout * 2.0

            except socket.error:
                raise

        # we didn't receive an ack, raise the timeout exception
        raise socket.timeout


    def recvfrom(self, bufsize=4096):
        # check if there is already data waiting from a transaction
        if self._received_data is not None:
            data = self._received_data
            addr = self._received_addr

            self._received_data = None
            self._received_addr = None

            return data, addr

        else:
            raise InvalidOperationException("recvfrom() but no ack from server")

    def close(self):
        self._sock.close()

    def shutdown(self, how=socket.SHUT_RDWR):
        try:
            self._sock.shutdown(how)

        except socket.error:
            pass


class TransparentClientSocket(ClientSocket):
    def sendto(self, data, address):
        # build data packet
        packet = Packet(data=data)

        self._sock.settimeout(self._initial_timeout)
        
        self._sock.sendto(packet.pack(), address)  

    def recvfrom(self, bufsize=4096):
        ack_data, host = self._sock.recvfrom(4096)

        # parse ack
        ack_msg = Packet().unpack(ack_data)

        if ack_msg.version != ack_msg.VERSION:
            raise InvalidPacketException(packet)

        elif not ack_msg.server:
            raise InvalidPacketException(packet)

        elif ack_msg.ack_request:
            raise InvalidPacketException(packet)

        return ack_msg.payload, host


class ServerSocket(object):

    def __init__(self, addr_family=None, sock_type=None):
        self.__sock = socket.socket(socket.AF_INET,
                                    socket.SOCK_DGRAM)

        self.__ack_host = None
        self.__ack_packet = None

    def bind(self, address):
        self.__sock.bind(address)

    def setsockopt(self, level, optname, value):
        self.__sock.setsockopt(level, optname, value)

    def settimeout(self, seconds):
        self.__sock.settimeout(seconds)

    def getsockname(self):
        return self.__sock.getsockname()

    def sendto(self, data="", address=None):
        # check if there is an ack packet to send
        if self.__ack_packet:

            # attach data to packet
            self.__ack_packet.payload = data

            # send packet
            self.__sock.sendto(self.__ack_packet.pack(), self.__ack_host)

            # delete ack packet and host
            self.__ack_host = None
            self.__ack_packet = None

        else:
            raise InvalidOperationException("sendto() but no message from client")

    def recvfrom(self, bufsize=4096):
        # receive packet and host address
        try:
            data, host = self.__sock.recvfrom(bufsize)

            # parse packet
            packet = Packet().unpack(data)

            # check packet for errors
            if packet.version != packet.VERSION:
                raise InvalidPacketException(packet)

            elif packet.server:
                raise InvalidPacketException(packet)

            elif packet.ack:
                raise InvalidPacketException(packet)

            # build and save ack packet
            self.__ack_packet = Packet(server=True,
                                       ack_request=False,
                                       ack=True,
                                       id=packet.id,
                                       data=packet.payload)

            # save client host
            self.__ack_host = host

            return packet.payload, host

        except socket.timeout:
            raise socket.timeout


class EchoServer(object):
    def __init__(self, address):
        self.__sock = ServerSocket()
        self.__sock.bind(address)

    def serve_forever(self):
        while True:
            data, host = self.__sock.recvfrom()
            self.__sock.sendto(data)

class InvalidOperationException(Exception):
    def __init__(self, value=None):
        self.value = value

    def __str__(self):
        return repr(self.value)

class InvalidPacketException(Exception):
    def __init__(self, value=None):
        self.value = value

    def __str__(self):
        return repr(self.value)


if __name__ == '__main__':

    c = ClientSocket()


    c.sendto("Jeremy", ("192.168.2.233", 1234))

    print c.recvfrom()



# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

import crcmod
import struct
from sapphire.common import util

crc_func = crcmod.predefined.mkCrcFun('crc-aug-ccitt')

@util.memoize
def kv_string_hash(s):
    crc0 = crc_func(s)
    crc1 = crc_func(''.join(reversed([chr(ord(a) ^ 0x15) for a in s])))

    h = (crc0 << 16) | crc1

    # convert to signed 32 bit int
    return struct.unpack('<l', struct.pack('<L', h))[0]



# <license>
# 
#     This file is part of the Sapphire Operating System.
# 
#     Copyright (C) 2013-2018  Jeremy Billheimer
# 
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# </license>

import time
import threading
import logging
import inspect
from Queue import Queue, Empty

class MsgQueueEmptyException(Exception):
    pass

class PositionalArgumentsNotSupported(Exception):
    pass


class Ribbon(threading.Thread):

    __lock = threading.Lock()
    __ribbons = []

    @classmethod
    def shutdown(cls, timeout=10.0):
        logging.info("Ribbon: shutting down all ribbons")

        with cls.__lock:
            for r in cls.__ribbons:
                r.stop()

        for i in xrange(int(timeout)):
            alive = False

            with cls.__lock:
                for r in cls.__ribbons:
                    if r.is_alive():
                        alive = True
                        break

            if alive:
                time.sleep(1.0)

            else:
                break

        with cls.__lock:
            for r in cls.__ribbons:
                if r.is_alive():
                    logging.error("Ribbon: %s failed to shut down" % (r.name))

    @classmethod
    def count(cls):
        return len(cls.__ribbons)

    def __init__(self,
                 name=None,
                 initialize_func=None,
                 clean_up_func=None,
                 loop_func=None,
                 auto_start=True,
                 queue=None,
                 **kwargs):
        super(Ribbon, self).__init__()

        if name:
            self.name = name

        self._stop_event = threading.Event()

        if initialize_func:
            self.initialize = initialize_func

        if clean_up_func:
            self.clean_up = clean_up_func

        if loop_func:
            self.loop = loop_func

        if queue:
            self.queue = queue

        else:
            self.queue = Queue()

        self.daemon = True

        with self.__lock:
            self.__ribbons.append(self)

        self._kwargs = kwargs

        try:
            # build dict of default kwargs for method
            argspec = inspect.getargspec(self.initialize)

            method_args = argspec.args[1:]
            defaults = argspec.defaults
            method_kwargs = {}
            for i in xrange(len(method_args)):
                try:
                    method_kwargs[method_args[i]] = defaults[i]

                except IndexError:
                    # this occurs if we pass a non-keyword argument
                    raise PositionalArgumentsNotSupported

            # now, override method kwargs
            for k, v in self._kwargs.iteritems():
                if k in method_kwargs:
                    method_kwargs[k] = v

            self.initialize(**method_kwargs)

            if auto_start:
                self.start()

        except Exception as e:
            logging.exception("Ribbon: %s failed to initialize with: %s" % (self.name, e))
            raise

    def start(self):
        super(Ribbon, self).start()

    def initialize(self, **kwargs):
        pass

    def clean_up(self):
        pass

    def loop(self):
        time.sleep(1.0)

    def run(self):
        # append thread ID to name
        self.name += '.%d' % (self.ident)

        logging.info("Ribbon: %s started" % (self.name))

        while not self._stop_event.is_set():
            try:
                self.loop()

            except Exception as e:
                logging.exception("Ribbon: %s unexpected exception: %s" % (self.name, e))
                time.sleep(2.0)


        try:
            self.clean_up()

        except Exception as e:
            logging.exception("Ribbon: %s failed to clean up with: %s" % (self.name, e))

        with self.__lock:
            self.__ribbons.remove(self)

        logging.info("Ribbon: %s stopped" % (self.name))

    def wait(self, timeout):
        self._stop_event.wait(timeout)

    def join(self, timeout=10.0):
        super(Ribbon, self).join(timeout)

        if self.is_alive():
            logging.info("Ribbon: %s join timed out" % (self.name))

    def stop(self):
        if not self._stop_event.is_set():
            logging.info("Ribbon: %s shutting down" % (self.name))

            self._stop_event.set()

    def post_msg(self, msg):
        self.queue.put(msg)

    def recv_msg(self, timeout=None):
        if timeout == None:
            q_timeout = 4.0
        else:
            q_timeout = timeout

        while not self._stop_event.is_set():
            try:
                return self.queue.get(timeout=q_timeout)

            except Empty:
                if timeout:
                    raise MsgQueueEmptyException

    def recv_all_msgs(self, timeout=None):
        if timeout == None:
            q_timeout = 4.0
        else:
            q_timeout = timeout

        msgs = []

        while not self._stop_event.is_set():
            try:
                msgs = [self.queue.get(timeout=q_timeout)]
                break

            except Empty:
                if timeout:
                    raise MsgQueueEmptyException

        qsize = self.queue.qsize()

        for i in xrange(qsize):
            msgs.append(self.queue.get())

        return msgs

def wait_for_signal():
    try:
        while True:
            time.sleep(1.0)

    except KeyboardInterrupt:
        pass

